@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2020 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -56,6 +56,9 @@ NOEXPORT void local_start(CLI *);
 NOEXPORT void remote_start(CLI *);
 NOEXPORT void ssl_start(CLI *);
 NOEXPORT void session_cache_retrieve(CLI *);
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT void print_tmp_key(SSL *s);
+#endif
 NOEXPORT void print_cipher(CLI *);
 NOEXPORT void transfer(CLI *);
 NOEXPORT int parse_socket_error(CLI *, const char *);
@@ -100,6 +103,18 @@ void *
     size_t stack_size=c->opt->stack_size;
 #endif
 
+#ifdef USE_FORK
+    /* do not use signal pipe in child processes */
+    signal(SIGCHLD, SIG_IGN); /* ignore dead children */
+    signal(SIGHUP, SIG_DFL);
+    signal(SIGUSR1, SIG_DFL);
+    signal(SIGUSR2, SIG_DFL);
+    signal(SIGPIPE, SIG_IGN); /* ignore broken pipe */
+    signal(SIGTERM, SIG_DFL);
+    signal(SIGQUIT, SIG_DFL);
+    signal(SIGINT, SIG_DFL);
+#endif /* USE_FORK */
+
     /* make sure c->thread_* values are initialized */
     CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_THREAD_LIST]);
     CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
@@ -485,6 +500,7 @@ NOEXPORT void remote_start(CLI *c) {
 
 NOEXPORT void ssl_start(CLI *c) {
     int i, err;
+    SSL_SESSION *sess;
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
     int unsafe_openssl;
 #endif /* OpenSSL version < 1.1.0 */
@@ -601,11 +617,17 @@ NOEXPORT void ssl_start(CLI *c) {
         throw_exception(c, 1);
     }
     print_cipher(c);
-    if(SSL_session_reused(c->ssl))
-        print_session_id(SSL_get_session(c->ssl));
-    /* SSL_SESS_CACHE_NO_INTERNAL_STORE prevented automatic caching */
-    if(!c->opt->option.client)
-        SSL_CTX_add_session(c->opt->ctx, SSL_get_session(c->ssl));
+    sess=SSL_get1_session(c->ssl);
+    if(sess) {
+        if(SSL_session_reused(c->ssl)) {
+            print_session_id(sess);
+        } else { /* a new session was negotiated */
+            /* SSL_SESS_CACHE_NO_INTERNAL_STORE prevented automatic caching */
+            if(!c->opt->option.client)
+                SSL_CTX_add_session(c->opt->ctx, sess);
+        }
+        SSL_SESSION_free(sess);
+    }
 }
 
 NOEXPORT void session_cache_retrieve(CLI *c) {
@@ -627,6 +649,42 @@ NOEXPORT void session_cache_retrieve(CLI *c) {
     CRYPTO_THREAD_unlock(stunnel_locks[LOCK_SESSION]);
 }
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT void print_tmp_key(SSL *s) {
+    EVP_PKEY *key;
+
+    if (!SSL_get_peer_tmp_key(s, &key)) {
+        sslerror("SSL_get_peer_tmp_key");
+        return;
+    }
+    switch (EVP_PKEY_id(key)) {
+    case EVP_PKEY_RSA:
+        s_log(LOG_INFO, "Peer temporary key: RSA, %d bits", EVP_PKEY_bits(key));
+        break;
+    case EVP_PKEY_DH:
+        s_log(LOG_INFO, "Peer temporary key: DH, %d bits", EVP_PKEY_bits(key));
+        break;
+#ifndef OPENSSL_NO_EC
+    case EVP_PKEY_EC:
+        {
+            EC_KEY *ec=EVP_PKEY_get1_EC_KEY(key);
+            int nid=EC_GROUP_get_curve_name(EC_KEY_get0_group(ec));
+            const char *cname=EC_curve_nid2nist(nid);
+            EC_KEY_free(ec);
+            if (cname == NULL)
+                cname=OBJ_nid2sn(nid);
+            s_log(LOG_INFO, "Peer temporary key: ECDH, %s, %d bits", cname, EVP_PKEY_bits(key));
+        }
+        break;
+#endif
+    default:
+        s_log(LOG_INFO, "Peer temporary key: %s, %d bits", OBJ_nid2sn(EVP_PKEY_id(key)),
+                   EVP_PKEY_bits(key));
+    }
+    EVP_PKEY_free(key);
+}
+#endif /* OpenSSL 1.1.1 or later */
+
 NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
     SSL_CIPHER *cipher;
 #ifndef OPENSSL_NO_COMP
@@ -645,6 +703,9 @@ NOEXPORT void print_cipher(CLI *c) { /* print negotiated cipher */
     s_log(LOG_INFO, "%s ciphersuite: %s (%d-bit encryption)",
         SSL_get_version(c->ssl), SSL_CIPHER_get_name(cipher),
         SSL_CIPHER_get_bits(cipher, NULL));
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+    print_tmp_key(c->ssl);
+#endif
 
 #ifndef OPENSSL_NO_COMP
     compression=SSL_get_current_compression(c->ssl);
@@ -664,7 +725,6 @@ NOEXPORT void transfer(CLI *c) {
     int has_pending=0, prev_has_pending;
 #endif
     int watchdog=0; /* a counter to detect an infinite loop */
-    ssize_t num;
     int err;
     /* logical channels (not file descriptors!) open for read or write */
     int sock_open_rd=1, sock_open_wr=1;
@@ -827,9 +887,9 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** send TLS close_notify alert */
         if(shutdown_wants_read || shutdown_wants_write) {
-            num=SSL_shutdown(c->ssl); /* send close_notify alert */
+            int num=SSL_shutdown(c->ssl); /* send close_notify alert */
             if(num<0) /* -1 - not completed */
-                err=SSL_get_error(c->ssl, (int)num);
+                err=SSL_get_error(c->ssl, num);
             else /* 0 or 1 - success */
                 err=SSL_ERROR_NONE;
             switch(err) {
@@ -837,16 +897,6 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_INFO, "SSL_shutdown successfully sent close_notify alert");
                 shutdown_wants_read=shutdown_wants_write=0;
                 break;
-            case SSL_ERROR_SYSCALL: /* socket error */
-                if(parse_socket_error(c, "SSL_shutdown"))
-                    break; /* a non-critical error: retry */
-                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                shutdown_wants_read=shutdown_wants_write=0;
-                break;
-            case SSL_ERROR_ZERO_RETURN: /* connection closed */
-                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                shutdown_wants_read=shutdown_wants_write=0;
-                break;
             case SSL_ERROR_WANT_WRITE:
                 s_log(LOG_DEBUG, "SSL_shutdown returned WANT_WRITE: retrying");
                 shutdown_wants_read=0;
@@ -860,6 +910,16 @@ NOEXPORT void transfer(CLI *c) {
             case SSL_ERROR_SSL: /* TLS error */
                 sslerror("SSL_shutdown");
                 throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* received a close_notify alert */
+                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                shutdown_wants_read=shutdown_wants_write=0;
+                break;
+            case SSL_ERROR_SYSCALL: /* socket error */
+                if(parse_socket_error(c, "SSL_shutdown"))
+                    break; /* a non-critical error: retry */
+                SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                shutdown_wants_read=shutdown_wants_write=0;
+                break;
             default:
                 s_log(LOG_ERR, "SSL_shutdown/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -868,7 +928,7 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** write to socket */
         if(sock_open_wr && sock_can_wr) {
-            num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
+            ssize_t num=writesocket(c->sock_wfd->fd, c->ssl_buff, c->ssl_ptr);
             switch(num) {
             case -1: /* error */
                 if(parse_socket_error(c, "writesocket"))
@@ -889,7 +949,7 @@ NOEXPORT void transfer(CLI *c) {
 
         /****************************** read from socket */
         if(sock_open_rd && sock_can_rd) {
-            num=readsocket(c->sock_rfd->fd,
+            ssize_t num=readsocket(c->sock_rfd->fd,
                 c->sock_buff+c->sock_ptr, BUFFSIZE-c->sock_ptr);
             switch(num) {
             case -1:
@@ -917,10 +977,10 @@ NOEXPORT void transfer(CLI *c) {
         /****************************** write to TLS */
         if((write_wants_read && ssl_can_rd) ||
                 (write_wants_write && ssl_can_wr)) {
+            int num=SSL_write(c->ssl, c->sock_buff, (int)(c->sock_ptr));
             write_wants_read=0;
             write_wants_write=0;
-            num=SSL_write(c->ssl, c->sock_buff, (int)(c->sock_ptr));
-            switch(err=SSL_get_error(c->ssl, (int)num)) {
+            switch(err=SSL_get_error(c->ssl, num)) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was written: ignore */
                     s_log(LOG_DEBUG, "SSL_write returned 0");
@@ -945,8 +1005,13 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_DEBUG,
                     "SSL_write returned WANT_X509_LOOKUP: retrying");
                 break;
+            case SSL_ERROR_SSL:
+                sslerror("SSL_write");
+                throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* a buffered close_notify alert */
+                /* fall through */
             case SSL_ERROR_SYSCALL: /* socket error */
-                if(num && parse_socket_error(c, "SSL_write"))
+                if(parse_socket_error(c, "SSL_write") && num) /* always log the error */
                     break; /* a non-critical error: retry */
                 /* EOF -> buggy (e.g. Microsoft) peer:
                  * TLS socket closed without close_notify alert */
@@ -959,14 +1024,6 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_INFO, "TLS socket closed (SSL_write)");
                 SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
-                s_log(LOG_INFO, "TLS closed (SSL_write)");
-                if(SSL_version(c->ssl)==SSL2_VERSION)
-                    SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                break;
-            case SSL_ERROR_SSL:
-                sslerror("SSL_write");
-                throw_exception(c, 1);
             default:
                 s_log(LOG_ERR, "SSL_write/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -978,10 +1035,10 @@ NOEXPORT void transfer(CLI *c) {
                 /* it may be possible to read some pending data after
                  * writesocket() above made some room in c->ssl_buff */
                 (read_wants_write && ssl_can_wr)) {
+            int num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, (int)(BUFFSIZE-c->ssl_ptr));
             read_wants_read=0;
             read_wants_write=0;
-            num=SSL_read(c->ssl, c->ssl_buff+c->ssl_ptr, (int)(BUFFSIZE-c->ssl_ptr));
-            switch(err=SSL_get_error(c->ssl, (int)num)) {
+            switch(err=SSL_get_error(c->ssl, num)) {
             case SSL_ERROR_NONE:
                 if(num==0) { /* nothing was read: ignore */
                     s_log(LOG_DEBUG, "SSL_read returned 0");
@@ -1004,8 +1061,17 @@ NOEXPORT void transfer(CLI *c) {
                 s_log(LOG_DEBUG,
                     "SSL_read returned WANT_X509_LOOKUP: retrying");
                 break;
+            case SSL_ERROR_SSL:
+                sslerror("SSL_read");
+                throw_exception(c, 1);
+            case SSL_ERROR_ZERO_RETURN: /* received a close_notify alert */
+                s_log(LOG_INFO, "TLS closed (SSL_read)");
+                if(SSL_version(c->ssl)==SSL2_VERSION)
+                    SSL_set_shutdown(c->ssl,
+                        SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+                break;
             case SSL_ERROR_SYSCALL:
-                if(num && parse_socket_error(c, "SSL_read"))
+                if(parse_socket_error(c, "SSL_read") && num) /* always log the error */
                     break; /* a non-critical error: retry */
                 /* EOF -> buggy (e.g. Microsoft) peer:
                  * TLS socket closed without close_notify alert */
@@ -1019,15 +1085,6 @@ NOEXPORT void transfer(CLI *c) {
                 SSL_set_shutdown(c->ssl,
                     SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
-                s_log(LOG_INFO, "TLS closed (SSL_read)");
-                if(SSL_version(c->ssl)==SSL2_VERSION)
-                    SSL_set_shutdown(c->ssl,
-                        SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
-                break;
-            case SSL_ERROR_SSL:
-                sslerror("SSL_read");
-                throw_exception(c, 1);
             default:
                 s_log(LOG_ERR, "SSL_read/SSL_get_error returned %d", err);
                 throw_exception(c, 1);
@@ -1146,22 +1203,15 @@ NOEXPORT void transfer(CLI *c) {
         shutdown_wants_read || shutdown_wants_write);
 }
 
-#ifdef __GNUC__
-#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
-#pragma GCC diagnostic push
-#endif /* __GNUC__>=4.6 */
-#if __GNUC__ >= 7
-#pragma GCC diagnostic ignored "-Wimplicit-fallthrough"
-#endif /* __GNUC__>=7 */
-#endif /* __GNUC__ */
-
     /* returns 0 on close and 1 on non-critical errors */
 NOEXPORT int parse_socket_error(CLI *c, const char *text) {
     switch(get_last_socket_error()) {
         /* http://tangentsoft.net/wskfaq/articles/bsd-compatibility.html */
     case 0: /* close on read, or close on write on WIN32 */
+        /* fall through */
 #ifndef USE_WIN32
     case EPIPE: /* close on write on Unix */
+        /* fall through */
 #endif
     case S_ECONNABORTED:
         s_log(LOG_INFO, "%s: Socket is closed", text);
@@ -1186,8 +1236,8 @@ NOEXPORT int parse_socket_error(CLI *c, const char *text) {
             s_log(LOG_INFO, "%s: Socket is closed (exec)", text);
             return 0;
         }
-        /* fall through */
 #endif
+        /* fall through */
     default:
         sockerror(text);
         throw_exception(c, 1);
@@ -1195,12 +1245,6 @@ NOEXPORT int parse_socket_error(CLI *c, const char *text) {
     }
 }
 
-#ifdef __GNUC__
-#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
-#pragma GCC diagnostic pop
-#endif /* __GNUC__>=4.6 */
-#endif /* __GNUC__ */
-
 NOEXPORT void auth_user(CLI *c) {
 #ifndef _WIN32_WCE
     struct servent *s_ent;    /* structure for getservbyname */
@@ -1469,9 +1513,13 @@ NOEXPORT SOCKET connect_remote(CLI *c) {
         if(!connect_init(c, c->connect_addr.addr[c->idx].sa.sa_family) &&
                 !s_connect(c, &c->connect_addr.addr[c->idx],
                     addr_len(&c->connect_addr.addr[c->idx]))) {
-            if(c->ssl)
-                idx_cache_save(SSL_get_session(c->ssl),
-                    &c->connect_addr.addr[c->idx]);
+            if(c->ssl) {
+                SSL_SESSION *sess=SSL_get1_session(c->ssl);
+                if(sess) {
+                    idx_cache_save(sess, &c->connect_addr.addr[c->idx]);
+                    SSL_SESSION_free(sess);
+                }
+            }
             print_bound_address(c);
             fd=c->fd;
             c->fd=INVALID_SOCKET;
@@ -1521,30 +1569,34 @@ NOEXPORT unsigned idx_cache_retrieve(CLI *c) {
     char *addr_txt;
 
     if(c->ssl && SSL_session_reused(c->ssl)) {
-        CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_ADDR]);
-        ptr=SSL_SESSION_get_ex_data(SSL_get_session(c->ssl),
-            index_session_connect_address);
-        if(ptr) {
-            len=addr_len(ptr);
-            memcpy(&addr, ptr, (size_t)len);
-            CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
-            /* address was copied, ptr itself is no longer valid */
-            for(i=0; i<c->connect_addr.num; ++i) {
-                if(addr_len(&c->connect_addr.addr[i])==len &&
-                        !memcmp(&c->connect_addr.addr[i],
-                            &addr, (size_t)len)) {
-                    addr_txt=s_ntop(&addr, len);
-                    s_log(LOG_INFO, "persistence: %s reused", addr_txt);
-                    str_free(addr_txt);
-                    return i;
+        SSL_SESSION *sess=SSL_get1_session(c->ssl);
+        if(sess) {
+            CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_ADDR]);
+            ptr=SSL_SESSION_get_ex_data(sess, index_session_connect_address);
+            if(ptr) {
+                len=addr_len(ptr);
+                memcpy(&addr, ptr, (size_t)len);
+                CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
+                SSL_SESSION_free(sess);
+                /* address was copied, ptr itself is no longer valid */
+                for(i=0; i<c->connect_addr.num; ++i) {
+                    if(addr_len(&c->connect_addr.addr[i])==len &&
+                            !memcmp(&c->connect_addr.addr[i],
+                                &addr, (size_t)len)) {
+                        addr_txt=s_ntop(&addr, len);
+                        s_log(LOG_INFO, "persistence: %s reused", addr_txt);
+                        str_free(addr_txt);
+                        return i;
+                    }
                 }
+                addr_txt=s_ntop(&addr, len);
+                s_log(LOG_INFO, "persistence: %s not available", addr_txt);
+                str_free(addr_txt);
+            } else {
+                CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
+                SSL_SESSION_free(sess);
+                s_log(LOG_NOTICE, "persistence: No cached address found");
             }
-            addr_txt=s_ntop(&addr, len);
-            s_log(LOG_INFO, "persistence: %s not available", addr_txt);
-            str_free(addr_txt);
-        } else {
-            CRYPTO_THREAD_unlock(stunnel_locks[LOCK_ADDR]);
-            s_log(LOG_NOTICE, "persistence: No cached address found");
         }
     }
 
@@ -1685,12 +1737,19 @@ NOEXPORT int connect_init(CLI *c, int domain) {
 }
 
 NOEXPORT int redirect(CLI *c) {
+    SSL_SESSION *sess;
+    void *ex_data;
+
     if(!c->opt->redirect_addr.names)
         return 0; /* redirect not configured */
     if(!c->ssl)
         return 1; /* TLS not established -> always redirect */
-    return !SSL_SESSION_get_ex_data(SSL_get_session(c->ssl),
-        index_session_authenticated);
+    sess=SSL_get1_session(c->ssl);
+    if(!sess)
+        return 1; /* no TLS session -> always redirect */
+    ex_data=SSL_SESSION_get_ex_data(sess, index_session_authenticated);
+    SSL_SESSION_free(sess);
+    return ex_data == NULL;
 }
 
 NOEXPORT void print_bound_address(CLI *c) {
