@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2020 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -42,6 +42,8 @@
 
 /**************************************** common constants */
 
+#define DEFAULT_SECURITY_LEVEL 2
+
 #define LIBWRAP_CLIENTS 5
 
 /* CPU stack size */
@@ -224,7 +226,6 @@ typedef int                 ssize_t;
 #define vsnprintf                   _vsnprintf
 #define strcasecmp                  _stricmp
 #define strncasecmp                 _strnicmp
-#define sleep(c)                    Sleep(1000*(c))
 
 #define get_last_socket_error()     WSAGetLastError()
 #define set_last_socket_error(e)    WSASetLastError(e)
@@ -316,6 +317,9 @@ typedef int SOCKET;
 #include <sys/select.h>     /* for aix */
 #endif
 #include <dirent.h>
+#ifdef HAVE_SYS_PARAM_H
+#include <sys/param.h>      /* MAXPATHLEN */
+#endif
 
 #if defined(HAVE_POLL) && !defined(BROKEN_POLL)
 #ifdef HAVE_POLL_H
