@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2020 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -99,6 +99,7 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
         signal(SIGCHLD, signal_handler); /* handle dead children */
         signal(SIGHUP, signal_handler); /* configuration reload */
         signal(SIGUSR1, signal_handler); /* log reopen */
+        signal(SIGUSR2, signal_handler); /* connections */
         signal(SIGPIPE, SIG_IGN); /* ignore broken pipe */
         if(signal(SIGTERM, SIG_IGN)!=SIG_IGN)
             signal(SIGTERM, signal_handler); /* fatal */
@@ -106,8 +107,20 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
             signal(SIGQUIT, signal_handler); /* fatal */
         if(signal(SIGINT, SIG_IGN)!=SIG_IGN)
             signal(SIGINT, signal_handler); /* fatal */
+#endif
+#ifdef USE_FORK
+        setpgid(0, 0); /* create a new process group if needed */
 #endif
         daemon_loop();
+#ifdef USE_FORK
+        s_log(LOG_NOTICE, "Terminating service processes");
+        signal(SIGCHLD, SIG_IGN);
+        signal(SIGTERM, SIG_IGN);
+        kill(0, SIGTERM); /* kill the whole process group */
+        while(wait(NULL)!=-1)
+            ;
+        s_log(LOG_NOTICE, "Service processes terminated");
+#endif
 #if !defined(__vms) && !defined(USE_OS2)
         delete_pid();
 #endif /* standard Unix */
@@ -126,6 +139,7 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
         tls_alloc(c, ui_tls, NULL);
         service_up_ref(&service_options);
         client_main(c);
+        client_free(c);
     }
     return 0;
 }
